# Copyright (c) 2022 The xlandsat developers.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""
Interpolation methods for scenes.
"""
import numpy as np
import scipy as sp


def interpolate_missing(scene, pixel_radius=20):
    """
    Fill missing values (NaNs) in a scene by cubic interpolation

    Each missing value is filled by interpolating the pixels within a
    neighboring region (controlled by ``pixel_radius``) using a piecewise cubic
    2D interpolator. Interpolation is done for each band in a scene separately.

    Note that this is mostly good if there are a few missing values, not large
    regions of the scene.

    Parameters
    ----------
    scene : :class:`xarray.Dataset`
        A Landsat scene, as read with :func:`xlandsat.load_scene`.
    pixel_radius : int
        Number of pixels to the east, west, south, and north of a missing value
        that will be used for interpolation. Smaller values make for faster
        interpolation but may lead to bad results if many missing values are
        grouped together.

    Returns
    -------
    filled_scene : :class:`xarray.Dataset`
        The scene with missing values filled in.
    """
    filled_scene = scene.copy(deep=True)
    rows, columns = np.meshgrid(
        np.arange(scene.northing.size),
        np.arange(scene.easting.size),
        indexing="ij",
    )
    for band in scene:
        values = filled_scene[band].values
        nans = np.isnan(values)
        for i, j in zip(rows[nans], columns[nans]):
            imin, imax = _search_range(i, pixel_radius, scene.northing.size)
            jmin, jmax = _search_range(j, pixel_radius, scene.easting.size)
            valid = ~np.isnan(values[imin:imax, jmin:jmax])
            interpolator = sp.interpolate.CloughTocher2DInterpolator(
                (
                    rows[imin:imax, jmin:jmax][valid],
                    columns[imin:imax, jmin:jmax][valid],
                ),
                values[imin:imax, jmin:jmax][valid],
            )
            values[i, j] = interpolator(i, j)
    return filled_scene


def _search_range(index, pixel_radius, dim_size):
    """
    Get a valid range around the index to search for interpolation data.
    Mostly used to avoid overflowing the image boundaries.
    """
    left, right = index - pixel_radius, index + pixel_radius
    if left < 0:
        left = 0
    if right > dim_size:
        right = dim_size
    return left, right



# Copyright (c) 2022 The xlandsat developers.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""
Test the download functions
"""
import pathlib

import pytest

from ..datasets import (
    fetch_brumadinho_after,
    fetch_brumadinho_before,
    fetch_liverpool,
    fetch_liverpool_panchromatic,
    fetch_manaus,
    fetch_momotombo,
    fetch_roraima,
)


@pytest.mark.parametrize("untar", [False, True], ids=["archive", "folder"])
@pytest.mark.parametrize(
    "fetcher",
    [
        fetch_brumadinho_after,
        fetch_brumadinho_before,
        fetch_liverpool,
        fetch_liverpool_panchromatic,
        fetch_manaus,
        fetch_roraima,
    ],
)
def test_fetching_functions(untar, fetcher):
    "Check that the download functions work"
    path = pathlib.Path(fetcher(untar=untar))
    assert path.exists()
    if untar:
        assert path.is_dir()
    else:
        assert not path.is_dir()


@pytest.mark.parametrize("untar", [False, True], ids=["archive", "folder"])
@pytest.mark.parametrize("level", [1, 2], ids=["L1", "L2"])
@pytest.mark.parametrize(
    "fetcher",
    [
        fetch_momotombo,
    ],
)
def test_fetching_functions_levels(untar, level, fetcher):
    "Check that the download functions work"
    path = pathlib.Path(fetcher(untar=untar, level=level))
    assert path.exists()
    if untar:
        assert path.is_dir()
    else:
        assert not path.is_dir()



# Copyright (c) 2022 The xlandsat developers.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""
Functions for downloading and caching sample datasets.
"""
import pathlib

import pooch

from ._version import __version__

POOCH = pooch.create(
    path=pooch.os_cache("xlandsat"),
    base_url="https://github.com/compgeolab/xlandsat/raw/{version}/data/",
    version=__version__,
    version_dev="main",
    registry={
        # Brumadinho - after
        "LC08_L2SP_218074_20190130_20200829_02_T1-cropped.tar.gz": "md5:4ae61a2d7a8b853c727c0c433680cece",
        # Brumadinho - before
        "LC08_L2SP_218074_20190114_20200829_02_T1-cropped.tar.gz": "md5:d2a503c944bb7ef3b41294d44b77e98c",
        # Liverpool
        "LC08_L2SP_204023_20200927_20201006_02_T1-cropped.tar.gz": "md5:3c07e343ccf959be4e5dd5c9aca4e0a4",
        # Liverpool - Panchromatic
        "LC08_L1TP_204023_20200927_20201006_02_T1-cropped.tar.gz": "md5:7d43f8580b8e583d137a93f9ae51a73d",
        # Momotombo L2
        "LC08_L2SP_017051_20151205_20200908_02_T1-cropped.tar.gz": "md5:8cc2e4c15e65940a7152fc1c8b412aa9",
        # Momotombo L1
        "LC08_L1TP_017051_20151205_20200908_02_T1-cropped.tar.gz": "md5:112d42e7adf709ac3a1179bbeedded6d",
        # Roraima
        "LC08_L2SP_232056_20151004_20200908_02_T1-cropped.tar.gz": "md5:f236a8b024aa4a4c62bee294d3bd737f",
        # Manaus
        "LC09_L2SP_231062_20230723_20230802_02_T1-cropped.tar.gz": "md5:ffe2003e665dc7a1a3155011f700a61d",
    },
)


def _fetch(fname, untar):
    """
    Fetch a file and handle untaring the archive if requested.
    """
    if untar:
        processor = pooch.Untar()
    else:
        processor = None
    path = POOCH.fetch(
        fname,
        processor=processor,
    )
    if untar:
        # Get the folder name in case we unpacked the archive
        path = pathlib.Path(path[0]).parent
    return path


def fetch_brumadinho_after(untar=False):
    """
    Download a sample scene from after the Brumadinho tailings dam disaster

    This is a cropped version of a Landsat 8 scene from 2019/01/30. It was
    taken after the `Brumadinho tailings dam in Brazil
    <https://en.wikipedia.org/wiki/Brumadinho_dam_disaster>`__ collapsed,
    flooding a whole region.

    The scene was downloaded from `USGS Earth Explorer
    <https://earthexplorer.usgs.gov/>`__. Original data are in the public
    domain and are redistributed here in accordance with the `Landsat Data
    Distribution Policy
    <https://www.usgs.gov/media/files/landsat-data-distribution-policy>`__.

    Source: https://doi.org/10.6084/m9.figshare.21665630
    (`CC0 <https://creativecommons.org/publicdomain/zero/1.0/>`__)

    Parameters
    ----------
    untar : bool
        If True, unpack the tar archive after downloading and return a path to
        the folder containing the unpacked files instead. Default is False.

    Returns
    -------
    path : str
        The path to the downloaded `.tar` file that contains the scene.
    """
    return _fetch("LC08_L2SP_218074_20190130_20200829_02_T1-cropped.tar.gz", untar)


def fetch_brumadinho_before(untar=False):
    """
    Download a sample scene from before the Brumadinho tailings dam disaster

    This is a cropped version of a Landsat 8 scene from 2019/01/14. It was
    taken before the `Brumadinho tailings dam in Brazil
    <https://en.wikipedia.org/wiki/Brumadinho_dam_disaster>`__ collapsed,
    flooding a whole region.

    The scene was downloaded from `USGS Earth Explorer
    <https://earthexplorer.usgs.gov/>`__. Original data are in the public
    domain and are redistributed here in accordance with the `Landsat Data
    Distribution Policy
    <https://www.usgs.gov/media/files/landsat-data-distribution-policy>`__.

    Source: https://doi.org/10.6084/m9.figshare.21665630
    (`CC0 <https://creativecommons.org/publicdomain/zero/1.0/>`__)

    Parameters
    ----------
    untar : bool
        If True, unpack the tar archive after downloading and return a path to
        the folder containing the unpacked files instead. Default is False.

    Returns
    -------
    path : str
        The path to the downloaded `.tar` file that contains the scene.
    """
    return _fetch(
        "LC08_L2SP_218074_20190114_20200829_02_T1-cropped.tar.gz",
        untar,
    )


def fetch_liverpool(untar=False):
    """
    Download a sample scene from the city of Liverpool, UK

    This is a cropped version of a Landsat 8 scene from 2020/09/27. It was
    taken on a virtually cloud-free day and shows the Mersey river delta and
    some off-shore wind turbines.

    The scene was downloaded from `USGS Earth Explorer
    <https://earthexplorer.usgs.gov/>`__. Original data are in the public
    domain and are redistributed here in accordance with the `Landsat Data
    Distribution Policy
    <https://www.usgs.gov/media/files/landsat-data-distribution-policy>`__.

    Source: https://doi.org/10.6084/m9.figshare.22041353
    (`CC0 <https://creativecommons.org/publicdomain/zero/1.0/>`__)

    Parameters
    ----------
    untar : bool
        If True, unpack the tar archive after downloading and return a path to
        the folder containing the unpacked files instead. Default is False.

    Returns
    -------
    path : str
        The path to the downloaded `.tar` file that contains the scene.
    """
    return _fetch(
        "LC08_L2SP_204023_20200927_20201006_02_T1-cropped.tar.gz",
        untar,
    )


def fetch_liverpool_panchromatic(untar=False):
    """
    Download a sample panchromatic band from the city of Liverpool, UK

    This is a cropped version of the panchromatic band from a Landsat 8 Level 1
    scene from 2020/09/27. It was taken on a virtually cloud-free day and shows
    the Mersey river delta and some off-shore wind turbines.

    The scene was downloaded from `USGS Earth Explorer
    <https://earthexplorer.usgs.gov/>`__. Original data are in the public
    domain and are redistributed here in accordance with the `Landsat Data
    Distribution Policy
    <https://www.usgs.gov/media/files/landsat-data-distribution-policy>`__.

    Source: https://doi.org/10.6084/m9.figshare.22041353
    (`CC0 <https://creativecommons.org/publicdomain/zero/1.0/>`__)

    Parameters
    ----------
    untar : bool
        If True, unpack the tar archive after downloading and return a path to
        the folder containing the unpacked files instead. Default is False.

    Returns
    -------
    path : str
        The path to the downloaded `.tar` file that contains the scene.
    """
    return _fetch(
        "LC08_L1TP_204023_20200927_20201006_02_T1-cropped.tar.gz",
        untar,
    )


def fetch_momotombo(untar=False, level=2):
    """
    Download a sample scene from the December 2015 Momotombo volcano eruption

    This is a cropped version of a Landsat 8 scene from 2015/12/05. It was
    taken during the December 2015 eruption of `Momotombo volcano
    <https://en.wikipedia.org/wiki/Momotombo>`__, Nicaragua.

    The scene was downloaded from `USGS Earth Explorer
    <https://earthexplorer.usgs.gov/>`__. Original data are in the public
    domain and are redistributed here in accordance with the `Landsat Data
    Distribution Policy
    <https://www.usgs.gov/media/files/landsat-data-distribution-policy>`__.

    Source: https://doi.org/10.6084/m9.figshare.21931089.v3
    (`CC0 <https://creativecommons.org/publicdomain/zero/1.0/>`__)

    Parameters
    ----------
    untar : bool
        If True, unpack the tar archive after downloading and return a path to
        the folder containing the unpacked files instead. Default is False.
    level : int
        Which level of data to load. Default is to load Level 2 surface
        reflectance data. Can also be Level 1 top-of-the-atmosphere
        reflectance.

    Returns
    -------
    path : str
        The path to the downloaded `.tar` file that contains the scene.
    """
    if level not in {1, 2}:
        raise ValueError(f"Invalid data level '{level}'. Must be 1 or 2.")
    if level == 2:
        data = _fetch(
            "LC08_L2SP_017051_20151205_20200908_02_T1-cropped.tar.gz",
            untar,
        )
    else:
        data = _fetch(
            "LC08_L1TP_017051_20151205_20200908_02_T1-cropped.tar.gz",
            untar,
        )
    return data


def fetch_roraima(untar=False):
    """
    Download a sample scene from Mount Roraima surrounded by clouds

    Roraima is a *tepui* located in the junction of Brazil, Guyana, and
    Venezuela. It's famous for the near-constant cloud coverage.

    This is a cropped version of a Landsat 8 scene from 2015/10/04, which is
    one of the rare relatively cloud-free scenes available.

    The scene was downloaded from `USGS Earth Explorer
    <https://earthexplorer.usgs.gov/>`__. Original data are in the public
    domain and are redistributed here in accordance with the `Landsat Data
    Distribution Policy
    <https://www.usgs.gov/media/files/landsat-data-distribution-policy>`__.

    Source: https://doi.org/10.6084/m9.figshare.24143622.v1
    (`CC0 <https://creativecommons.org/publicdomain/zero/1.0/>`__)

    Parameters
    ----------
    untar : bool
        If True, unpack the tar archive after downloading and return a path to
        the folder containing the unpacked files instead. Default is False.

    Returns
    -------
    path : str
        The path to the downloaded `.tar` file that contains the scene.
    """
    return _fetch("LC08_L2SP_232056_20151004_20200908_02_T1-cropped.tar.gz", untar)


def fetch_manaus(untar=False):
    """
    Download a sample scene from Manaus, Brazil

    Manaus is located in the Brazilian Amazon. The scene shows a part of the
    city and the meeting of the Solimões and Negro rivers to form the Amazon
    river.

    This is a cropped version of a Landsat 9 scene from 2023/07/23, during
    the annual Amazon river floods.

    The scene was downloaded from `USGS Earth Explorer
    <https://earthexplorer.usgs.gov/>`__. Original data are in the public
    domain and are redistributed here in accordance with the `Landsat Data
    Distribution Policy
    <https://www.usgs.gov/media/files/landsat-data-distribution-policy>`__.

    Source: https://doi.org/10.6084/m9.figshare.24167235.v1
    (`CC0 <https://creativecommons.org/publicdomain/zero/1.0/>`__)

    Parameters
    ----------
    untar : bool
        If True, unpack the tar archive after downloading and return a path to
        the folder containing the unpacked files instead. Default is False.

    Returns
    -------
    path : str
        The path to the downloaded `.tar` file that contains the scene.
    """
    return _fetch("LC09_L2SP_231062_20230723_20230802_02_T1-cropped.tar.gz", untar)



# Copyright (c) 2022 The xlandsat developers.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""
Operations to enhance composites. Basically wrappers for skimage.exposure.
"""
import numpy as np
import skimage.color
import skimage.exposure


def equalize_histogram(composite, kernel_size=None, clip_limit=0.01):
    """
    Adaptive histogram equalization for a composite

    Use this function to enhance the contrast of a composite when there are a
    few very dark or very light patches that dominate the color range. Use this
    instead of rescaling intensity (contrast stretching) to try to preserve
    some detail in the light/dark patches.

    If the composite has an alpha channel (transparency), it will be copied to
    the output intact.

    .. warning::

        Results can be very bad if there are missing values (NaNs) in the
        composite. Use :func:`xlandsat.interpolate_missing` on the scene first
        (before creating the composite) if that is the case.

    Parameters
    ----------
    composite : :class:`xarray.DataArray`
        A composite, as generated by :func:`xlandsat.composite`.
    kernel_size : int or None
        The size of region used in the algorithm. See
        :func:`skimage.exposure.equalize_adapthist` for details and defaults.
    clip_limit : float
        Clipping limit, normalized between 0 and 1 (higher values give more
        contrast).

    Returns
    -------
    equalized_composite : :class:`xarray.DataArray`
        The composite after equalization, scaled back to the range of the
        original composite.

    Notes
    -----

    This function first converts the composite from the RGB color space to the
    `HSV color space <https://en.wikipedia.org/wiki/HSL_and_HSV>`__. Then, it
    applies :func:`skimage.exposure.equalize_adapthist` to the values
    (intensity) channel. Finally, the composite is converted back into the RGB
    color space.
    """
    result = composite.copy(deep=True)
    # Make sure the input is in the 0-255 range for the color space transform
    hsv = skimage.color.rgb2hsv(
        skimage.exposure.rescale_intensity(result.values[:, :, :3], out_range="uint8")
    )
    hsv[:, :, 2] = skimage.exposure.equalize_adapthist(
        hsv[:, :, 2],
        kernel_size=kernel_size,
        clip_limit=clip_limit,
    )
    result.values[:, :, :3] = skimage.exposure.rescale_intensity(
        skimage.color.hsv2rgb(hsv),
        out_range=str(composite.values.dtype),
    )
    return result


def adjust_l1_colors(composite, percentile=0.1):
    """
    Adjust the colors in an RGB composite from Level 1 data

    Corrects the balance of the red, green, and blue bands in an RGB (true
    color) composite made from Level 1 data (without atmospheric correction).
    This is **not as accurate as atmospheric correction** but can lead to nicer
    images in places where the atmospheric correction causes artifacts.

    **Do not use the output for calculating indices.**

    Parameters
    ----------
    composite : :class:`xarray.DataArray`
        A composite, as generated by :func:`xlandsat.composite`.
    percentile : float
        The percentile range to use for the intensity rescaling. Will use the
        ``percentile`` as the lower bound and ``100 - percentile`` for the
        upper bound. Default is 0.1%.

    Returns
    -------
    adjusted_composite : :class:`xarray.DataArray`
        The composite after color adjustment, scaled back to the range of the
        original composite.

    Notes
    -----

    The correction raises each channel to the power of 1/2.2 (inverse of the
    sRGB gamma function) and then rescales the transformed intensity to the
    given percentile range.
    """
    output = composite.copy()
    for i, channel in enumerate(["red", "green", "blue"]):
        scaled = composite.sel(channel=channel).values ** (1 / 2.2)
        vmin, vmax = np.percentile(scaled, (percentile, 100 - percentile))
        output.values[:, :, i] = skimage.exposure.rescale_intensity(
            scaled,
            in_range=(vmin, vmax),
            out_range=str(composite.values.dtype),
        )
    return output
